; framework sound driver for SCI0
; interacts with interpreter but does not play anything
; assembles with NASM (see http://sourceforge.net/projects/nasm/):
;    nasm nosnd.asm -o nosnd.drv -f bin

; Copyright (c) 2000, 2002, 2003 by
;    Ravi Iyengar [ravi.i@softhome.net]
;    Andy Hefner
;
; This library is free software; you can redistribute it and/or
; modify it under the terms of the GNU Lesser General Public
; License as published by the Free Software Foundation; either
; version 2.1 of the License, or (at your option) any later version.
;
; This library is distributed in the hope that it will be useful,
; but WITHOUT ANY WARRANTY; without even the implied warranty of
; MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
; Lesser General Public License for more details.
;
; You should have received a copy of the GNU Lesser General Public
; License along with this library; if not, write to the Free Software
; Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA


[BITS 16]

; Drivers are flatform binaries that get loaded at the start of
; a segment. When a driver function needs to be executed, the
; interpreter does a far call to offset 0. The first thing must
; be a jump to a DriverInterface function because the interpreter
; expects certain data starting at offset 3, and there isn't
; space to fit in anything else.

jmp near DriverInterface

; identifiers for a Sierra sound driver
db 0                   ; uncertain
dd 0x87654321          ; Sierra driver id
db 1                   ; sound driver id (0=display, 4=keyboard)

; CHANGEME:
; a string identifying the driver
db 8, 'nosnddrv'

; CHANGEME:
; a string with a description of the device
db 38, 'No sound / Framework driver BETA 0.5.1'



;--------------------------------------------------------------------------
; misc macros and defines

; CHANGEME:
; device information returned by GetDeviceInfo
%define  INFO_PATCH  0xFFFF  ; patch resource number (0xFFFF means none)
%define  INFO_POLY   0       ; maximum polyphony

; CHANGEME:
; the play flag for this device (used in the sound header)
%define  PLAY_FLAG   0

; playstate constants
%define  STOPPED     0
%define  PLAYING     1
%define  SEEKING     2

; offsets in the sound info structure
%define  SND_RESPTR  8       ; heap pointer to far pointer to resource data
%define  SND_UNK     10      ; zeroed to fix a CB fadeout bug
%define  SND_POS     12      ; current play offset in sound resource
%define  SND_STATE   16      ; 1 = valid, 3 = invalid   (possibly others)
%define  SND_SIGNAL  22      ; the Sound object's signal property
%define  SND_VOLUME  24      ; global sound volume



;--------------------------------------------------------------------------
; data area
; Sierra's drivers put all their data space in the beginning, though
;    there's no reason why it has to be that way.

ptrparam:      dw 0          ; saves the original si parameter to DriverInterface
playstate:     db 0          ; 0 = not playing sound, 1 = playing sound, 2 = seeking sound
fadeticks:     db 0          ; number of ticks to next volume decrement, 0 = not fading
fadevolume:    db 0          ; the volume setting during fades
playflags:     dw 0          ; specifies whether or not each channel is used by this device
waitcount:     dw 0          ; number of ticks left before next event
sndresptr:     dw 0          ; pointer to sound resource
looppoint:     dw 0          ; sound loop position
cumcue:        dw 0          ; the cumulative cue
status:        db 0          ; saves the last status for running status mode, holds the current status
pssndreset:    db 0          ; boolean describing whether PauseSound should reset the sound
seekto:        dw 0          ; the end point for SeekSound
seekcue:       dw 0          ; the desired sound cue at the end of SeekSound



;--------------------------------------------------------------------------
; export area
; the interpreter interacts with the driver through these functions


; Export Lookup Table
; used by DriverInterface to call exported functions by number
; Exports must be even numbers because offsets are 2 bytes each.
;    Later interpreters used odd numbers for export functions as
;    well and multiplied by 2 to get the offset in this table.

ExportTable:
   Export0:    dw GetDeviceInfo
   Export2:    dw InitDevice
   Export4:    dw ShutdownDevice
   Export6:    dw LoadSound
   Export8:    dw DoSoundEvents
   Export10:   dw SetVolume
   Export12:   dw FadeOut
   Export14:   dw StopSound
   Export16:   dw PauseSound
   Export18:   dw SeekSound


; DriverInterface
; Saves registers and calls export
; Entry: bp = export number
;        ds = heap segment
;        si = heap pointer where applicable
; Exit: ax, cx = return values where applicable
;       data at ds:si modified where applicable

DriverInterface:
   pushf
   push dx
   push bx
   push sp
   push bp
   push si
   push di
   push es
   push ds
   mov [cs:ptrparam], si
   call near word [cs:ExportTable+bp]
   pop ds
   pop es
   pop di
   pop si
   pop bp
   pop sp
   pop bx
   pop dx
   popf
   retf


; GetDeviceInfo
; Called at startup to determine max poly and which patch
;    resource to load.
; Entry: nothing
; Exit: ax = patch number (0xFFFF if no patch)
;       cx = maximum polyphony

GetDeviceInfo:
   mov ax, INFO_PATCH
   mov cx, INFO_POLY
   retn


; InitDevice
; Called to load the patch if applicable and initialize the device
; Entry: ds:si points to patch resource structure
; Exit: cs:ax points to the area of the driver segment which can be overwritten
;       cx = 0

; The interpreter wants to be able to use portions of the sound driver's
; memory that won't be used by the driver. InitDevice returns a pointer
; through ax that tells the interpreter where it can begin overwriting things
; as necessary. Polite drivers should put InitDevice and GetDeviceInfo at the
; end, and allow the interpreter to overwrite these functions since they'll
; only be used once anyway. If you choose to do this, unlist destroyed
; exports from the export lookup table. I don't think that doing so is
; strictly necessary, but Sierra's drivers do it.

InitDevice:
   ; the next three lines can be removed if there is no patch to load
   mov bx, [si+8]
   mov es, [bx+2]
   mov di, [bx]              ; es:di points to patch data if applicable

   ; CHANGEME: Initialize sound hardware
   ; If the device can't be opened, set ax to 0xFFFF
   ; instead of temp_area

   mov ax, temp_area
   xor cx, cx
   retn


; ShutdownDevice
; Called to close the device when the interpreter exits. There
;    is no guarantee that StopSound will have been called on a
;    currently playing sound.
; Entry: ds:si points to sound resource structure
; Exit: nothing

ShutdownDevice:
   ; CHANGEME: Silence the hardware
   ; CHANGEME: Close the sound hardware
   retn


; LoadSound
; Called before a sound starts playing so the driver has a chance
;    to load the header and perform initialization work.
; Entry: ds:si points to sound resource structure
; Exit: ax = 3 on unrecognized header byte, 1 on success

LoadSound:
   mov word [cs:playflags], 0x8000 ; channel 15 (the control channel) should always be enabled

   mov bx, [si+SND_RESPTR]
   mov es, [bx+2]
   mov di, [bx]              ; es:di points to sound data

   xor dx, dx                ; dx will be added to di to compensate for missing channel 15 info
   mov cx, 16
   mov al, byte [es:di]      ; check for the presence of a PCM sample
   inc di
   cmp al, 0
   je load_header
   cmp al, 2
   je fifteen_chan

   mov ax, 3                 ; unplayable / invalid resource
   mov word [si+SND_STATE], ax
   retn

   fifteen_chan:
   dec cx                     ; if there is a PCM sample, the header only has info for 15 channels
   mov dx, 2                  ; compensates for missing channel 15 info

   load_header:
   mov bx, 1

   load_header_loop:
   mov al, byte [es:di]      ; al = initial number of voices on this channel
   inc di
   mov ah, byte [es:di]      ; ah = play flag for this channel
   inc di
   test ah, PLAY_FLAG        ; check for playflag
   jz continue_load

   or word [cs:playflags], bx ; enable channel

   ; CHANGEME: Initialize the MIDI channel

   continue_load:
   shl bx, 1
   loop load_header_loop

   add di, dx                 ; dx = 2 if there was only init data for 15 channels

   call getwaitcount

   mov ax, 1
   mov word [si+SND_STATE], ax  ; success loading sound
   mov word [si+SND_POS], di    ; set the position to the start of sound data
   mov word [si+SND_SIGNAL], 0  ; set the signal to 0
   mov word [cs:cumcue], 127    ; initialize the cumulative cue
   mov word [cs:looppoint], 33  ; initialize the loop point to the top
   mov word [cs:pssndreset], 0  ; PauseSound does not reset by default
   mov byte [cs:fadeticks], 0   ; there is no fade in progress
   mov byte [cs:playstate], PLAYING ; we're playing this sound now
   
   mov cx, word [si+SND_VOLUME]
   shl cx, 2                 ; we use a range of 0-63 for internal global volume
   ; CHANGEME: Set the global volume to cx

   retn


; DoSoundEvents
; Called each MIDI tick (60 times per second) to actually play the
;    sound resource
; Entry: ds:si points to sound resource structure
; Exit: nothing

StatusCallTable:             ; call table for handling status messages
   dw StopNote
   dw PlayNote
   dw KeyPressure
   dw ControlChange
   dw PatchChange
   dw ChanPressure
   dw PitchWheel

ParamCount: db 2, 2, 2, 2, 1, 1, 2 ; number of parameters used by each status

DoSoundEvents:
   mov bx, [si+SND_RESPTR]
   mov es, [bx+2]
   mov di, [bx]              ; es:di points to sound resource
   mov word [cs:sndresptr], di
   add di, word [si+SND_POS] ; get the current sound resource position

   call fader
   cmp byte [cs:playstate], STOPPED
   jne do_frame
   retn                      ; the sound isn't playing anymore, so quit

   do_frame:                 ; the loop to do all events in this frame (all events with delta time = 0)

   cmp word [cs:waitcount], 0 ; check if there's a new event
   je next_event
   dec word [cs:waitcount]
   jmp end_frame

   next_event:
   xor cx, cx
   mov cl, byte [es:di]
   inc di
   test cl, 0x80             ; check for running status mode
   jnz check_status
   mov cl, byte [cs:status]  ; repeat the last status
   dec di                    ; rewind di for re-reading the last byte as a parameter

   check_status:
   mov dl, cl                ; dl = full status byte (status and channel)
   mov byte [cs:status], cl
   mov al, cl
   shr al, 4                 ; al = status
   and cl, 0x0F              ; cl = channel

   cmp al, 0x0F
   je sys_message            ; check for a system message before we check the channel

   mov bx, 1
   shl bx, cl
   test word [cs:playflags], bx ; test to see if this channel is enabled
   jnz do_status

   mov bl, al                ; al is still the status
   sub bl, 8
   xor bh, bh
   mov bl, [cs:ParamCount+bx]
   add di, bx
   call getwaitcount
   jmp do_frame              ; this channel is to be ignored, so skip this status

   do_status:
   mov bl, al                ; al is still the status
   sub bl, 8
   xor bh, bh
   shl bx, 1
   call word [cs:StatusCallTable+bx] ; call the handler for this particular status
   call getwaitcount
   jmp do_frame

   sys_message:
   ; FIXME: We assume that this is 0xFC, but we should really check
   mov word [si+SND_SIGNAL], 0xFFFF ; the sound is done - set the signal appropriately
   xor bx, bx
   mov [cs:waitcount], bx
   mov bx, word [si+SND_RESPTR]
   mov di, word [bx]
   add di, word [cs:looppoint] ; set es:di to appropriate loop point

   end_frame:
   sub di, word [cs:sndresptr]
   mov word [si+SND_POS], di ; update the sound position
   retn


; SetVolume
; Called when the user changes the global sound volume (Ctrl-V in the game)
; Entry: ds:si points to sound resource structure
; Exit: nothing

SetVolume:
   mov cx, [si+SND_VOLUME]   ; cx = volume level (0-15)
   shl cx, 2                 ; we use a range of 0-63 for internal global volume
   ; CHANGEME: Set the global volume to cx
   retn


; FadeOut
; Called to slowly fade and stop the currently playing sound
; Entry: ds:si points to sound resource structure
; Exit: nothing

FadeOut:
   mov cx, [si+SND_VOLUME]   ; cx = volume level (0-15)
   shl cx, 2                 ; we use a range of 0-63 for internal global volume
   test cx, cx
   jz f_stop                 ; just stop the sound now if the volume starts at 0

   mov [cs:fadevolume], cl
   mov byte [cs:fadeticks], 5
   retn

   f_stop:
   mov word [si+SND_SIGNAL], 0xFFFF ; the sound is done - set the signal appropriately
   ; CHANGEME: Silence the hardware
   mov byte [cs:playstate], STOPPED
   retn


; StopSound
; Stops playback by turning off sound output. Nothing needs to be done to
;   prevent further MIDI events from occurring because the interpreter will
;   simply stop calling DoSoundEvents.
; Entry: ds:si points to sound resource structure
; Exit: the sound's signal property is set to -1

StopSound:
   ; CHANGEME: Silence the hardware
   mov word [si+SND_SIGNAL], 0xFFFF
   mov word [cs:playstate], STOPPED
   retn


; PauseSound
; Pauses playback by turning off sound output. Nothing needs to be done to
;   prevent further MIDI events from occurring because the interpreter will
;   simply stop calling DoSoundEvents until playback resumes. If CC4C has been
;   set, the sound position should be reset. Otherwise it should be left alone.
; Entry: ds:si points to sound resource structure
; Exit: nothing

PauseSound:
   ; CHANGEME: Silence the hardware

   cmp byte [cs:pssndreset], 0
   je exitpausesound

   mov ax, word [cs:looppoint] ; reset sound position to loop point
   mov word [si+SND_POS], ax

   mov bx, [si+SND_RESPTR]
   mov es, [bx+2]
   mov di, [bx]              ; es:di points to sound resource
   add di, ax                ; move forward to the loop position
   call getwaitcount         ; get a new waitcount for when the sound resumes

   exitpausesound:
   retn


; SeekSound
; Called when a sound needs to be quickly set to a certain point (e.g. on
;    restore, when one sound interrupts another, and so on)
; Entry: ds:si points to sound resource structure
; Exit: nothing

SeekSound:
   mov ax, word [si+SND_POS] ; this is where we need to seek to
   mov word [cs:seekto], ax
   mov ax, word [si+SND_SIGNAL] ; this is the what the cue should be
   call LoadSound
   mov byte [cs:playstate], SEEKING
   cmp ax, 1
   je do_seek
   retn

   do_seek:
   call DoSoundEvents
   mov bx, word [si+SND_POS]
   cmp bx, word [cs:seekto]
   jb do_seek                ; make sure we're at the right sound position
   mov bx, word [si+SND_SIGNAL]
   cmp bx, word [cs:seekcue]
   jne do_seek               ; make sure we're at the right cue position

   mov byte [cs:playstate], PLAYING
   retn



;--------------------------------------------------------------------------
; status functions

; on entry for each of these functions:
;   al = status
;   cl = channel
;   dl = full status byte (status and channel)
;   es:di points to parameters for the status
; on exit:
;   es:di points to the delta time until the next sound element


; we can group these together because we treat them the same - real drivers may have to separate them
KeyPressure:
PitchWheel:
   ; CHANGEME: However your driver wants to treat these
   add di, 2
   ret


StopNote:
   cmp byte [cs:playstate], PLAYING
   je stop_note

   add di, 2                 ; we're just seeking, don't play anything
   ret

   stop_note:
   mov bh, byte [es:di]      ; bh = note number
   mov bl, byte [es:di+1]    ; bl = velocity
   add di, 2
   ; CHANGEME: Stop the note
   ret


PlayNote:
   cmp byte [cs:playstate], PLAYING
   je play_note

   add di, 2                 ; we're just seeking, don't play anything
   ret

   play_note:
   mov bh, byte [es:di]      ; bh = note number
   mov bl, byte [es:di+1]    ; bl = velocity
   add di, 2
   ; CHANGEME: Play the note
   ret


ControlChange:
   mov bh, byte [es:di]      ; bh = control number
   mov bl, byte [es:di+1]    ; bl = control value
   add di, 2
   cmp bh, 0x60
   jne not_CC60h

   xor bh, bh
   add bx, word [cs:cumcue]
   mov word [si+SND_SIGNAL], bx ; set the new signal
   mov word [cs:cumcue], bx  ; update our cumulative cue variable
   ret

   not_CC60h:
   cmp bh, 0x4C
   jne not_CC4Ch

   mov byte [cs:pssndreset], bl ; set the PauseSound reset control value
   ret

   not_CC4Ch:                ; there are other special controls, but we don't care right now
   ; CHANGEME: Standard control change
   ret


PatchChange:
   mov bl, byte [es:di]      ; bl = patch
   inc di
   cmp cl, 15
   jne normal_patch_change

   cmp bl, 127
   je set_loop_point

   mov byte [si+SND_SIGNAL], bl ; this is a cue - set the signal property
   ret

   set_loop_point:           ; set the sound loop point
   push di
   sub di, word [cs:sndresptr]
   sub di, 2
   mov word [cs:looppoint], di
   pop di
   ret

   normal_patch_change:      ; this is a simple patch change
   ; CHANGEME: Standard patch change
   ret


ChanPressure:
   ; CHANGEME: However your driver wants to treat this
   add di, 1
   ret



;--------------------------------------------------------------------------
; misc functions


; getwaitcount
; retrieves the next delta time from the sound resource
; entry: es:di points to the current delta time value
; exit: [cs:waitcount] set, es:di points to the sound element following the delta time

getwaitcount:
   push ax
   push dx
   xor ax, ax
   xor dx, dx

   get_delta_time_loop:
   mov dl, byte [es:di]
   cmp dl, 0xF8
   jne add_dt

   mov dl, 0xF0              ; delta time extension

   add_dt:
   add ax, dx
   inc di
   cmp dl, 0xF0
   je get_delta_time_loop

   mov word [cs:waitcount], ax ; set waitcount
   pop dx
   pop ax
   ret


; fader
; checks for a fade in progress and acts accordingly
; entry: es:di points to the current position in the sound resource
; exit:

fader:
   cmp byte [cs:fadeticks], 0 ; check to see if there's a fade in progress
   je exit_fader

   dec byte [cs:fadeticks]
   cmp byte [cs:fadeticks], 0 ; check if we need to decrease the volume yet
   jne exit_fader

   dec byte [cs:fadevolume]
   cmp byte [cs:fadevolume], 0 ; just stop the sound if we're done fading
   je f_stopsnd

   mov byte [cs:fadeticks], 5
   push cx
   xor cx, cx
   mov cl, byte [cs:fadevolume]
   ; CHANGEME: Set the global volume to cx
   pop cx
   jmp exit_fader

   f_stopsnd:
   mov byte [cs:playstate], STOPPED
   push si
   mov si, [cs:ptrparam]
   mov word [si+SND_SIGNAL], 0xFFFF ; the sound is done - set the signal appropriately
   mov word [si+SND_UNK], 0  ; fixes fadeout bugs in CB
   ; CHANGEME: Silence the hardware
   pop si

   exit_fader:
   ret



;--------------------------------------------------------------------------
; data / functions that will only be used before or during initialization
temp_area:
